;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Installation instructions
;; 1. Put this file somewhere (if it's not already in rostools/rosemacs in your
;;    ros tree)
;; 2. (Optional) From emacs do M-x byte-compile followed by this file's full path 
;; 3. Add the following lines to ~/.emacs:
;;    (require 'rosemacs)
;;    (invoke-rosemacs)
;; 4. (Optional) Add the following line or equivalent to ~/.emacs to have
;;    quick access to added commands:
;;    (global-set-key "\C-x\C-r" ros-keymap)
;; 5. Start emacs from a shell in which the standard ROS variables are set, and
;;    the variable EMACSLOADPATH is set to include the location of this file
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Usage
;; 1. Directory tracking and tab completion for rosbash commands should now work 
;;    correctly in shell mode
;; 2. The commands {find|view}-ros-{file|message|service}, and 
;;    view-most-recent-ros-log for navigating the ros libraries are available.  
;;    Tab completion should work for all of them.
;; 3. Use ros-update-topic-list to make rosemacs update its list of topics, 
;;    and set-ros-topic-update-interval to do so periodically in the background.
;;    This will enable tab completion of ros topics in the shell and for commands
;;    such as echo-ros-topic.  Additionally, you can use add-hz-update to 
;;    define a list of topics for which the hz rate is tracked in the background,
;;    viewable using display-ros-topic-info.
;; 4. Also, if the ros-keymap was bound to some prefix in step 4 above, the 
;;    above commands are accessible with key sequences starting with that 
;;    prefix (press control-h after the prefix to see the list).  For example,
;;    \C-x\C-r\C-f would call find-ros-file.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(require 'shell)
(require 'cl)
(require 'warnings)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Parameters
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar ros-ignored-packages '("test_msgs" "test_roscpp_serialization_perf") "List of packages to ignore when searching")
(defvar ros-topic-timeout-rate 5 "Number of seconds before info from rostopic hz is considered out-of-date" )
(defvar ros-topic-display-update-interval 3 "Number of seconds between updates to the *rostopic* buffer (when it's visible)")
(defvar ros-topic-update-interval nil "Gap in seconds between calls to rostopic list (end of one call to beginning of next).  nil means never call.")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; State
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar ros-packages nil "Vector of ros packages")
(defvar ros-package-locations nil "Vector of directories containing the items in ros-packages")
(defvar ros-messages nil "Vector of ros messages")
(defvar ros-message-packages nil "Vector of packages corresponding to each ros message")
(defvar ros-services nil "Vector of ros services")
(defvar ros-service-packages nil "Vector of packages corresponding to each service")
(defvar ros-root (getenv "ROS_ROOT"))
(defvar ros-topics nil "Vector of current ros topics")
(defvar ros-topic-completions (vector) "Completion table for topics (ros-topics, without the initial /")
(defvar ros-topic-hertz-processes nil "Alist from topic name to process running rostopic hz on that topic")
(defvar ros-topic-publication-rates nil "Hash table from topic name to hertz rate of that topic")
(defvar ros-topic-last-hz-rate nil "Alist from topic name to last time we saw output from rostopic hz")
(defvar ros-topic-buffer nil "Holds the buffer *ros-topics* if it exists")
(defvar ros-hz-topic-regexps nil "If a topic name matches one of these, it is hz tracked")
(defvar ros-topic-timer nil "If non-nil, equals timer object used to schedule calls to rostopic list")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Preloading
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun ros-load-package-locations ()
  "Reload locations of ros packages by calling out to rospack list"
  (interactive)
  (with-temp-buffer
    (let ((l nil))
      (message "Calling rospack")
      (call-process "rospack" nil t nil "list")
      (goto-char (point-min)) 
      (message "Parsing rospack output")
      (let ((done nil))
	;; Loop over lines; each line contains a package and directory
	(while (not done)
	  (let ((p (point)))
	    ;; Search for string terminated by space
	    (setq done (not (re-search-forward "[[:space:]]" (point-max) t)))
	    (unless done
	      (let ((package (buffer-substring p (1- (point)))))
		(setq p (point))
		;; Search for following string terminated by newline
		(re-search-forward "\n")
		(let ((dir (buffer-substring p (1- (point)))))
		  (unless (member package ros-ignored-packages)
		    (push (cons package dir) l))))))))
      (let ((package-alist (sort* (vconcat l) (lambda (pair1 pair2) (string< (car pair1) (car pair2))))))
	(setq ros-packages (map 'vector #'car package-alist)
	      ros-package-locations (map 'vector #'cdr package-alist)
	      ros-messages nil
	      ros-message-packages nil
	      ros-services nil
	      ros-service-packages nil))
      (message "Done loading ROS package info"))))

(defun ros-messages-in-package (dir ext)
  "Return list of files in subdirectory ext/ of dir whose extension is .ext"
  (with-temp-buffer
    (let ((l nil)
	  (done nil)
	  (p nil))
      (call-process "ls" nil t nil (concat dir "/" ext "/"))
      (goto-char (point-min))
      (while (not done)
	(setq p (point))
	(setq done (not (re-search-forward "\\([^[:space:]]+\\)[[:space:]]+" (point-max) t)))
	(unless done
	  (let ((str (buffer-substring (match-beginning 1) (match-end 1))))
	  (let ((m (string-match (concat "\." ext "$") str)))
	    (when m
	      (push (substring str 0 m) l))))))
      l)))
      

(defun all-files-in-packages (ext)
  "Look in each package for files with a extension .ext in subdirectory ext/"
  (unless ros-package-locations
    (ros-load-package-locations))
  (let ((l nil))
    (dotimes-with-progress-reporter (i (length ros-package-locations)) (concat "Caching locations of ." ext " files: ")
      (let ((package (aref ros-packages i))
	    (dir (aref ros-package-locations i)))
	(dolist (m (ros-messages-in-package dir ext))
	  (push (cons m package) l))))
    (sort* (vconcat l) (lambda (pair1 pair2) (string< (car pair1) (car pair2))))))

(defun cache-ros-message-locations ()
  "Look in each package directory for .msg files"
  (let ((v (all-files-in-packages "msg")))
    (setq ros-messages (map 'vector #'car v)
	  ros-message-packages (map 'vector #'cdr v))))

(defun cache-ros-service-locations ()
  "Look in each package directory for .srv files"
  (let ((v (all-files-in-packages "srv")))
    (setq ros-services (map 'vector #'car v)
	  ros-service-packages (map 'vector #'cdr v))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Lookup
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun ros-package-dir (package)
  (unless ros-package-locations
    (ros-load-package-locations))
  (lookup-vectors package ros-packages ros-package-locations))

(defun ros-message-package (m)
  (unless ros-message-packages
    (cache-ros-message-locations))
  (lookup-vectors m ros-messages ros-message-packages))

(defun ros-service-package (m)
  (unless ros-service-packages
    (cache-ros-service-locations))
  (lookup-vectors m ros-services ros-service-packages))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; parsing
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun parse-ros-file-prefix (str)
  "Divide something of the form PACKAGE/DIRS/FILE-PREFIX into its three pieces.  Or, if it's just a package prefix, return just that."
  (if (string-match "\\([^/]+\\)\\(/.*\\)" str)
      (let ((package (match-string 1 str))
	    (path (match-string 2 str)))
	(if (string-match "\\(/.*/\\)\\([^/]*\\)" path)
	    (values package (match-string 1 path) (match-string 2 path))
	  (values package "/" (substring path 1))))
    (values str nil nil)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Completion
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(setq message-completor (dynamic-completion-table (lambda (str) (unless ros-messages (cache-ros-message-locations)) (bsearch-completions str ros-messages))))
(setq service-completor (dynamic-completion-table (lambda (str) (unless ros-services (cache-ros-service-locations)) (bsearch-completions str ros-services))))
(setq topic-completor (dynamic-completion-table (lambda (str) (bsearch-completions str ros-topic-completions))))
(setq package-completor 
      ;; Longer because it has to deal with the case of PACKAGE/PATH-PREFIX in addition to PACKAGE-PREFIX
      (dynamic-completion-table 
       (lambda (str) 
	 (unless ros-packages (ros-load-package-locations))
	 (multiple-value-bind (package dir-prefix dir-suffix) (parse-ros-file-prefix str)
	   (if dir-prefix
	       (let ((dir (concat (ros-package-dir package) dir-prefix)))
		 (let* ((files (directory-files dir nil (concat "^" dir-suffix)))
			(comps (all-completions dir-suffix files)))
		   (mapcar (lambda (comp) (concat package dir-prefix comp)) comps)))
	     (bsearch-completions package ros-packages))))))



(defun comint-dynamic-complete-ros-package ()
  ;; Like the above, except in the shell
  (unless ros-packages (ros-load-package-locations))
  (let ((prefix (comint-get-ros-package-prefix)))
    (when prefix 
      (multiple-value-bind (package dir-prefix dir-suffix) (parse-ros-file-prefix prefix)
	(if dir-prefix
	    (let ((dir (concat (ros-package-dir package) dir-prefix)))
	      (let ((completions (all-completions dir-suffix (directory-files dir nil (concat "^" dir-suffix)))))
		(comint-dynamic-simple-complete dir-suffix completions)
		(skip-syntax-backward " ")))
	  (progn
	  (comint-dynamic-simple-complete prefix (all-completions prefix package-completor))
	  (skip-syntax-backward " ")))))))

(defun comint-dynamic-complete-ros-topic ()
  (let ((prefix (comint-get-ros-topic-prefix)))
    (when prefix
      (comint-dynamic-simple-complete prefix (all-completions prefix topic-completor))
      (skip-syntax-backward " "))))


	

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Navigation commands
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun find-ros-file (package-name &optional dont-reload)
  "Open up the directory corresponding to PACKAGE-NAME in dired mode.  If used interactively, tab completion will work."
  (interactive (list (completing-read "Enter ros path: " package-completor) nil))
  (multiple-value-bind (package dir-prefix dir-suffix) (parse-ros-file-prefix package-name)
    (let* ((package-dir (ros-package-dir package))
	   (path (if dir-prefix (concat package-dir dir-prefix dir-suffix) package-dir)))
      (if path
	  (find-file path)
	(if dont-reload
	    (error "Did not find %s in the ros package list." package-name)
	  (progn
	  (lwarn '(rosemacs) :debug "Did not find %s.  Reloading ros package list and trying again..." package-name)
	  (ros-load-package-locations)
	  (find-ros-file package-name t)))))))

(defun view-ros-file (ros-file-name &optional dont-reload)
  "View (open in read-only mode with simpler editing commands — see emacs help) the file corresponding to ROS-FILE-NAME (in form packagename/filename).  If used interactively, tab completion will work."
  (interactive (list (completing-read "Enter ros path: " package-completor) nil))
  (multiple-value-bind (package dir-prefix dir-suffix) (parse-ros-file-prefix ros-file-name)
    (let* ((package-dir (ros-package-dir package))
	   (path (if dir-prefix (concat package-dir dir-prefix dir-suffix) package-dir)))
      (if path
	  (view-file-other-window path)
	(if dont-reload
	    (error "Did not find %s in the ros package list." ros-file-name)
	  (progn
	  (lwarn '(rosemacs) :debug "Did not find %s.  Reloading ros package list and trying again..." ros-file-name)
	  (ros-load-package-locations)
	  (view-ros-file ros-file-name t)))))))

(defun find-ros-message (message)
  "Open definition of a ros message.  If used interactively, tab completion will work."
  (interactive (list (completing-read 
		      (if (current-word t t)
			  (format "Enter message name (default %s): " (current-word t t))
			"Enter message name: ")
		      message-completor nil nil nil nil (current-word t t))))
  (let ((p (ros-message-package message)))
    (if p
	(let ((dir (ros-package-dir p)))
	  (if dir
	      (find-file (concat dir "/msg/" message ".msg"))
	    (error "Could not find directory corresponding to package %s" p)))
      (error "Could not find package for message %s" message))))

(defun find-ros-service (service)
  "Open definition of a ros service.  If used interactively, tab completion will work."
  (interactive (list (completing-read 
		      (if (current-word t t)
			  (format "Enter service name (default %s): " (current-word t t))
			"Enter service name: ")
		      service-completor nil nil (current-word t t))))
  (let ((p (ros-service-package service)))
    (if p
	(let ((dir (ros-package-dir p)))
	  (if dir
	      (find-file (concat dir "/srv/" service ".srv"))
	    (error "Could not find directory corresponding to package %s" p)))
      (error "Could not find package for service %s" service))))


(defun view-ros-message (message)
  "Open definition of a ros message in view mode.  If used interactively, tab completion will work."
  (interactive (list (completing-read
		      (if (current-word t t)
			  (format "Enter message name (default %s): " (current-word t t))
			"Enter message name: ")
		      message-completor nil nil nil nil (current-word t t))))
  (let ((p (ros-message-package message)))
    (if p
	(let ((dir (ros-package-dir p)))
	  (if dir
	      (view-file-other-window (concat dir "/msg/" message ".msg"))
	    (error "Could not find directory corresponding to package %s" p)))
      (error "Could not find package for message %s" message))))

(defun view-ros-service (service)
  "Open definition of a ros service in view mode.  If used interactively, tab completion will work."
  (interactive (list (completing-read 
		      (if (current-word t t)
			  (format "Enter service name (default %s): " (current-word t t))
			"Enter service name: ")
		      service-completor nil nil (current-word t t))))
  (let ((p (ros-service-package service)))
    (if p
	(let ((dir (ros-package-dir p)))
	  (if dir
	      (view-file-other-window (concat dir "/srv/" service ".srv"))
	    (error "Could not find directory corresponding to package %s" p)))
      (error "Could not find package for service %s" service))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Core
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun ros-core ()
  (interactive)
  (if (get-buffer "*roscore*")
      (switch-to-buffer (get-buffer "*roscore*"))
    (progn (start-process "roscore" (get-buffer-create "*roscore*") "roscore")
	   (message "roscore started"))))

(defun ros-set-master-uri (host port)
  "Set the master uri used by other commands (e.g. rostopic)"
  (interactive "sEnter master uri host: \nnEnter master uri port: ")
  (let ((uri (format "http://%s:%d" host port)))
    (setenv "ROS_MASTER_URI" uri)
    (message "Set ROS_MASTER_URI to %s" uri)))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; rostopic
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Top-level
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun ros-update-topic-list ()
  "Makes rosemacs call rostopic list and update its list of topics (ros-topics)"
  (interactive)

  (let ((ros-topic-update-interval 0)) ;; dynamic binding
    (ros-update-topic-list-internal)))

(defun set-ros-topic-update-interval (n)
  "Make rostopic list be called every n seconds starting now.  0 means never update."
  (interactive "nEnter rostopic update interval in seconds (0 means never update) : ")
  (cond
   ((= n 0)
    (setq ros-topic-update-interval nil))
   (t
    (setq ros-topic-update-interval n)
    (ros-update-topic-list-internal))))
  


(defun display-ros-topic-info ()
  "Display current ros topic info in *ros-topics* buffer"
  (interactive)
  (let ((buf (get-buffer-create "*ros-topics*")))
    (setq ros-topic-buffer buf)
    (switch-to-buffer buf)
    (update-ros-topic-buffer)))

(defun add-hz-update (topic-regexp)
  (interactive (list (completing-read "Enter topic name or regexp to track: " topic-completor)))

  ;; Asynchronously start re-gathering topic list, in case things have recently changed
  (ros-update-topic-list)
  
  (push topic-regexp ros-hz-topic-regexps)
  (dolist (topic ros-topics)
    (when (string-match topic-regexp topic)
      (unless (assoc topic ros-topic-last-hz-rate)
	(start-hz-tracker topic)))))
  

(defun remove-hz-update (topic-regexp)
  (interactive (list (completing-read "Enter regexp to stop tracking: " ros-hz-topic-regexps)))
  (setq ros-hz-topic-regexps (delete topic-regexp ros-hz-topic-regexps))
  (dolist (pair ros-topic-last-hz-rate)
    (let ((topic (car pair)))
      (when (string-match topic-regexp topic)
	(stop-hz-tracker topic)))))
  


(defun echo-ros-topic (topic)
  "Create a new buffer in which rostopic echo is done on the given topic (read interactively, with tab-completion)"
  (interactive (list (let ((word (current-word)))
		       (completing-read
			(if word
			(format "Enter topic name (default %s): " word)
			"Enter topic name: ")
			topic-completor nil nil nil nil word))))
  (let* ((topic-full-name (if (string-match "^/" topic) topic (concat "/" topic)))
	 (buffer-name (concat "*rostopic:" topic-full-name "*"))
	 (process (start-process buffer-name buffer-name "rostopic" "echo" topic-full-name)))
    (switch-to-buffer (process-buffer process))
    (help-mode)
    (ros-topic-echo-mode)
    ))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Internal
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun interrupt-ros-topic-echo ()
  (interactive)
  (interrupt-process))

(defun kill-ros-topic-buffer ()
  (interactive)
  (kill-buffer nil))

(defvar ros-topic-echo-keymap (make-sparse-keymap))
(define-key ros-topic-echo-keymap "k" 'interrupt-ros-topic-echo)
(define-key ros-topic-echo-keymap "q" 'kill-ros-topic-buffer)

(define-minor-mode ros-topic-echo-mode 
  "Mode used for rostopic echo.  

k kills the process (sends SIGINT).
q kills the buffer and process"
  :init-value nil
  :lighter " ros-topic-echo"
  :keymap ros-topic-echo-keymap
  (message "ros-topic-echo-mode: k to stop, q to quit"))


(defun ros-update-topic-list-internal ()
  (lwarn '(rosemacs) :debug "calling rostopic list")
  (let* ((b (generate-new-buffer "rostopic"))
	 (p (start-process "rostopic" b "rostopic" "list")))
    (set-process-sentinel p 'schedule-parse-rostopic-list)))


(defun update-ros-topic-buffer ()
  "Use the current value of ros-topic related variables to reset the contents of the *ros-topics* buffer, if it's visible"
  (when (and ros-topic-buffer (get-buffer-window ros-topic-buffer))
    (let ((old-point (when (equal (current-buffer) ros-topic-buffer) (point))))
      (save-excursion
	(set-buffer ros-topic-buffer)
	(erase-buffer)
	(princ (format "Master uri: %s\n" (getenv "ROS_MASTER_URI")) ros-topic-buffer)
	(princ (format "Topics updated every %s seconds\n" ros-topic-update-interval) ros-topic-buffer)


	(when ros-topic-publication-rates
	  (princ (format "\nHz-tracked topics:\n") ros-topic-buffer)
	  (dolist (topic ros-topics)
	    (let ((rate-pair (assoc topic ros-topic-publication-rates)))
	      (when rate-pair
		(let ((rate (cdr rate-pair))
		      (diff (- (second (current-time)) (or (cdr (assoc topic ros-topic-last-hz-rate)) 0.0))))
		  (if rate
		      (if (> diff ros-topic-timeout-rate)
			  (princ (format " %s : %s (as of %s seconds ago)" topic rate diff) ros-topic-buffer)
			(princ (format " %s : %s" topic rate) ros-topic-buffer))
		    (princ (format " %s : not yet known" topic) ros-topic-buffer)))
		(terpri ros-topic-buffer))))
	  (terpri ros-topic-buffer))

	(princ (format "\nTopics:\n") ros-topic-buffer)
	(dolist (topic ros-topics)
	  (princ (format " %s" topic) ros-topic-buffer)
	  (terpri ros-topic-buffer)))
      (when old-point
	(goto-char old-point)))))
    




(defun schedule-parse-rostopic-list (process event)
  "Use the output of rostopic list to recompute the list of published topics, and add and remove new/removed topics.  Kill the buffer of the rostopic process at the end."
  (lwarn '(rosemacs) :debug "rostopic list returned.")
  (progn
    (sit-for (round (* .75 (or ros-topic-update-interval 0.0))))
    (when ros-topic-timer (cancel-timer ros-topic-timer))
    (setq ros-topic-timer (run-with-idle-timer (round (* .25 (or ros-topic-update-interval 0.0))) nil 'parse-rostopic-list process event))))


(defun parse-rostopic-list (process event)
  (lwarn '(rosemacs) :debug "Parsing rostopic list")
  (unwind-protect
      (if (string-match "finished" event)
	  (save-excursion
	    (unless (active-minibuffer-window)
	      (switch-to-buffer (process-buffer process))
	      (goto-char (point-min))
	      (lwarn '(rosemacs) :debug "1")
	      (let ((start (re-search-forward "Published topics:" nil t)))
		(when start
		  (let ((end (re-search-forward "Subscribed topics:")))
		    (setq end (or end (point-max)))
		    (lwarn '(rosemacs) :debug "3")
	    
		    ;; At this point, start and end point to the beginning and end of the published topics
		    (let ((done nil) (current-topics nil))
		      (goto-char start)
		      (while (not done)
			(let ((pos (re-search-forward "^[[:space:]]*\\*[[:space:]]+\\([^[:space:]]+\\)" end t)))
			  (if pos
			      (push (match-string 1) current-topics)
			    (setq done t))))
		      (setq current-topics (sort* current-topics 'string<))
			
		      (destructuring-bind (added deleted) (diffs ros-topics current-topics)
			(lwarn '(rosemacs) :debug "added topics : %s" added)
			(dolist (topic added)
			  (add-ros-topic topic))
			(dolist (topic deleted)
			  (remove-ros-topic topic))
			)))))))
	      
	(lwarn '(rosemacs) :warning "rostopic process received unexpected event %s." event))

    (lwarn '(rosemacs) :debug "Done parsing rostopic list")
    (when ros-topic-update-interval
      (ros-update-topic-list-internal))
    (kill-buffer (process-buffer process))))






(defun regenerate-ros-topic-completions ()
  "Use the current value of ros-topics to regenerate the variable ros-topic-completions.  Include both full topic names (foo/bar/baz) and local names (baz) for each one."
  (setq ros-topic-completions 
	(sort* (vconcat
		(mapcar (lambda (topic)
			  (string-match "^/*\\(.*\\)" topic)
			  (match-string 1 topic))
			ros-topics))
		'string<)))

(defun remove-ros-topic (topic)
  "Remove this topic and all associated entries from topic list, completion list, hertz processes, publication rates"
  (message "removing ros topic %s" topic)
  (stop-hz-tracker topic) 
  (setq ros-topics (delete topic ros-topics))
  (regenerate-ros-topic-completions)
  )

(defun stop-hz-tracker (topic)
  (let ((pair (assoc topic ros-topic-hertz-processes)))
    (when pair (kill-buffer (process-buffer (cdr pair)))))
  (setq ros-topic-hertz-processes (delete-if (lambda (pair) (equal (car pair) topic)) ros-topic-hertz-processes)
	ros-topic-publication-rates (delete-if (lambda (pair) (equal (car pair) topic)) ros-topic-publication-rates)
	ros-topic-last-hz-rate (delete-if (lambda (pair) (equal (car pair) topic)) ros-topic-last-hz-rate)))



(defun set-ros-topic-hz (topic rate)
  "Set hertz rate of topic.  Also, update the last-published-hertz-rate timestamp of the topic"
  (let ((rate-pair (assoc topic ros-topic-publication-rates)))
    (if rate-pair
	(setf (cdr rate-pair) rate)
      (push (cons topic rate) ros-topic-publication-rates)))
  (let ((last-time-pair (assoc topic ros-topic-last-hz-rate))
	(time-in-seconds (second (current-time))))
    (if last-time-pair
	(setf (cdr last-time-pair) time-in-seconds)
      (push (cons topic time-in-seconds) ros-topic-last-hz-rate)))
  (lwarn '(rosemacs) :debug "Updated hz for topic %s" topic))

(defun ros-topic-hz-filter (proc string)
  "Given the hertz process and string containing output from it, update the entry for the corresponding topic's publication rate"
  (let ((pair (rassoc proc ros-topic-hertz-processes)))
    (if pair
	(let ((hz (ros-topic-extract-hz string))
	      (topic (car pair)))
	  (cond 
	   ((eql hz 'not-published) (remove-ros-topic topic))
	   (hz (set-ros-topic-hz topic hz))))
      (lwarn '(rosemacs) :warning "Unexpectedly could not find topic corresponding to process %s" (process-name proc)))))


(defun ros-topic-extract-hz (string)
  "Given a string, if it contains something that looks like a hz reading, return the first such one, else if it contains error, return the string error, else if not published yet, return the number -1, else return nil"
  ;; This will not work well if the process filter ends up being called too frequently, in which case no one call will match the regexp.
  ;; Empirically, that does not seem to happen.
  (let ((case-fold-search t))
    (cond
     ((string-match "error" string) "error")
     ((string-match "average rate: \\([0-9]+\\.[0-9]*\\)" string) (match-string 1 string))
     ((string-match "does not appear to be published yet" string) 'not-published))))


(defun start-hz-tracker (topic)
  (let* ((name (concat "rostopic-hz-" topic))
	 (proc (start-process name name "rostopic" "hz" topic)))
    (push (list topic) ros-topic-last-hz-rate)
    (push (list topic) ros-topic-publication-rates)
    (let ((old-proc-pair (assoc topic ros-topic-hertz-processes)))
      (if old-proc-pair
	  (progn
	    (kill-buffer (process-buffer (cdr old-proc-pair)))
	    (setf (cdr old-proc-pair) proc))
	(push (cons topic proc) ros-topic-hertz-processes)))
    (set-process-filter proc 'ros-topic-hz-filter)))

(defun satisfies-hz-regexps (topic)
  (some (lambda (regexp) (string-match regexp topic)) ros-hz-topic-regexps))

(defun add-ros-topic (topic)
  "Post: topic is added to the list of topics and related bookkeeping done: list of completions is regenerated, hertz process is started and added to list, publication-rate entry is added to list"
  (let ((pair (assoc topic ros-topic-publication-rates)))
    (when pair (setq ros-topic-publication-rates (delete pair ros-topic-publication-rates))))
  (when (satisfies-hz-regexps topic)
    (start-hz-tracker topic))
  (push topic ros-topics)
  (setq ros-topics (sort* ros-topics 'string<))
  (regenerate-ros-topic-completions)
  )



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Shell mode
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun ros-directory-tracker (str)
  "Keep buffer current directory up-to-date in the presence of roscd.  This is analogous to shell-directory-tracker in shell.el though not as thorough about special cases.  It will not work with variable substitution, any sort of program control, etc - only for simple commands and pipelines of the same.  If it gets confused, do M-x shell-resync-dirs."
  ;; Might be simpler to just call shell-resync-dirs at some point

  ;; skip whitespace
  (let ((start (progn (string-match 
		       (concat "^" shell-command-separator-regexp) str)
		      (match-end 0)))
	end cmd arg1)
    (while (string-match shell-command-regexp str start)

      (setq end (match-end 0)
	    cmd (comint-arguments (substring str start end) 0 0)
	    arg1 (comint-arguments (substring str start end) 1 1))
      (when arg1 (setq arg1 (shell-unquote-argument arg1)))
      
      (cond ((string-match "^roscd\\([[:space:]]\\|$\\)" cmd)
	     (if (string-match "\\([^/]*\\)/\\(.*\\)" arg1)
		 (let ((package (match-string 1 arg1))
		       (subdir (match-string 2 arg1)))
		   (message "Package is %s, subdir is %s" package subdir)
		   (let ((dir (ros-package-dir package)))
		     (if dir
			 (shell-process-cd (concat dir "/" subdir))
		       (lwarn '(rosemacs) :warning "Unable to find directory of ros package %s." arg1))))
	       (let ((dir (ros-package-dir arg1)))
		 (if dir
		     (shell-process-cd dir)
		   (lwarn '(rosemacs) :warning "Unable to find directory of ros package %s." arg1))))))
      ;; TODO deal with popd

      (setq start (progn (string-match shell-command-separator-regexp str end)
			 (match-end 0))))))



(defun ros-emacs-last-word ()
  (let ((end (point)))
    (skip-syntax-backward "w_.()")
    (buffer-substring-no-properties (point) end)))
  
(defvar *ros-commands-starting-with-package* '("roscd" "rosmake" "rosrun" "rospd"))

(defun comint-get-ros-package-prefix ()
  (ros-command-prefix *ros-commands-starting-with-package*))

(defun ros-command-prefix (commands)
  (save-excursion
    (block match-block
      (let ((arg (ros-emacs-last-word)))
	(skip-syntax-backward " ")
	(dolist (cmd commands nil)
	  (when (string-equal cmd (buffer-substring-no-properties (- (point) (length cmd)) (point)))
	    (return-from match-block arg)))))))

(defun comint-get-ros-topic-prefix ()
  (save-excursion
    (let ((arg (ros-emacs-last-word)))
      (skip-syntax-backward " ")
      (ros-emacs-last-word)
      (skip-syntax-backward " ")
      (let ((start (- (point) 8)))
	(when (and (>= start 0) (string-equal "rostopic" (buffer-substring-no-properties start (point))))
	  arg)))))

      
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Keymap
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar ros-keymap (make-sparse-keymap))
(define-key ros-keymap "\C-f" 'find-ros-file)
(define-key ros-keymap "f" 'view-ros-file)
(define-key ros-keymap "\C-m" 'find-ros-message)
(define-key ros-keymap "m" 'view-ros-message)
(define-key ros-keymap "\C-s" 'find-ros-service)
(define-key ros-keymap "s" 'view-ros-service)
(define-key ros-keymap "\C-r" 'ros-load-package-locations)
(define-key ros-keymap "\C-u" 'set-ros-topic-update-interval)
(define-key ros-keymap "u" 'ros-update-topic-list)
(define-key ros-keymap "\C-c" 'ros-core)
(define-key ros-keymap "\C-t" 'display-ros-topic-info)
(define-key ros-keymap "t" 'echo-ros-topic)
(define-key ros-keymap "h" 'add-hz-update)
(define-key ros-keymap "H" 'remove-hz-update)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Invoking the mode
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun set-rosemacs-shell-hooks ()
  (add-hook 'comint-input-filter-functions 'ros-directory-tracker nil t)
  (add-hook 'comint-dynamic-complete-functions 'comint-dynamic-complete-ros-package nil t)
  (add-hook 'comint-dynamic-complete-functions 'comint-dynamic-complete-ros-topic nil t))


(defun invoke-rosemacs ()
  (interactive)
  (add-hook 'shell-mode-hook 'set-rosemacs-shell-hooks)
  (set-ros-topic-update-interval 0)
  (run-at-time t ros-topic-display-update-interval 'update-ros-topic-buffer))

    


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Internal
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun bsearch-completions (str completions)
  "str is a string, completions is a sorted vector of strings.  Return list of strings in completions that str is a prefix of."
  (let ((num-completions (length completions)))
    (unless (or (= num-completions 0)
		(string< (aref completions (1- num-completions)) str))
    (let ((i 0)
	  (j (1- num-completions)))
      (while (< (1+ i) j)
	(let ((k (floor (+ i j) 2)))
	  (if (string< str (aref completions k))
	      (setq j k)
	    (setq i k))))

      (when (not (is-prefix str (aref completions i)))
	(incf i))
      ;; Postcondition: completions of str, if they exist, begin at i
      
      (let ((returned-completions nil))
	(while (and (< i (length completions)) (is-prefix str (aref completions i)))
	  (push (aref completions i) returned-completions)
	  (incf i))
	returned-completions)))))

(defun is-prefix (str1 str2)
  (let ((m (length str1)))
    (eq t (compare-strings str1 0 m str2 0 m))))

(defun lookup-vectors (str v1 v2)
  (let ((i (position str v1 :test #'string-equal)))
    (when i
      (aref v2 i)))) 

(defun diffs (l1 l2)
  "Given two sorted lists of strings, return a list with 1) l2 - l1 2) l1 - l2"
  (let ((added nil) (deleted nil) (remaining1 l1) (remaining2 l2))
    (while (or remaining1 remaining2)
      (cond
       ((or (null remaining1) (and remaining2 (string< (car remaining2) (car remaining1))))
	(push (car remaining2) added)
	(setq remaining2 (cdr remaining2)))
       ((or (null remaining2) (and remaining1 (string< (car remaining1) (car remaining2))))
	(push (car remaining1) deleted)
	(setq remaining1 (cdr remaining1)))
       (t (setq remaining1 (cdr remaining1)
		remaining2 (cdr remaining2)))))
    (lwarn '(rosemacs) :debug "Diffs of %s and %s are %s and %s" l1 l2 added deleted)
    (list added deleted)))

    
(provide 'rosemacs)